"""Text utilities."""

import re

WORD_START = re.compile(r"(?<!^)(?=[A-Z])")


def to_snakecase(string: str):
    """Convert a string to snake_case."""
    return WORD_START.sub("_", string).lower()


def speech_duration(
    text: str, words_per_second: float = 3.3, min_duration: float = 1.0
) -> float:
    """Calculate the duration of a speech given the text."""
    return max(len(text.split()) / words_per_second, min_duration)



"""A simple immediate mode GUI for Pygame."""

import contextlib
import inspect
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Any

import pygame
from pygame.event import Event
from pygame.locals import K_BACKSPACE
from pygame.locals import K_TAB
from pygame.locals import KEYDOWN
from pygame.locals import KMOD_SHIFT

from pygskin.rect import add_padding
from pygskin.rect import get_rect_attrs

CLICKABLE = 1
EDITABLE = 2
FOCUSABLE = 4
SCROLLABLE = 8
HAS_BORDER = 16
HAS_BACKGROUND = 32
HAS_SHADOW = 64


Align = Enum("Align", "left center right")
VAlign = Enum("VAlign", "top middle bottom")


@dataclass
class Widget:
    """Immediate mode GUI widget."""

    value: Any = None
    flags: int = 0
    rect: pygame.Rect | None = None
    triggered: bool = False
    type: str | None = None
    id: str | None = None
    classes: list[str] = field(default_factory=list)
    pseudo_classes: set[str] = field(default_factory=set)


def _draw_widget(widget: Widget, surface: pygame.Surface, **style) -> None:
    if "focus" in widget.pseudo_classes:
        _draw_widget_focus(surface, widget, **style)
    _draw_widget_background(surface, widget, **style)
    _draw_widget_border(surface, widget, **style)
    _draw_widget_text(surface, widget, **style)


def label(text: str | list[str], **kwargs) -> Widget:
    """Create a label widget."""
    return Widget(text, type="label", **kwargs)


def button(text: str | list[str], **kwargs) -> Widget:
    """Create a button widget."""
    return Widget(
        text,
        CLICKABLE | FOCUSABLE | HAS_BORDER | HAS_BACKGROUND,
        type="button",
        **kwargs,
    )


def textfield(text: list[str], **kwargs) -> Widget:
    """Create a textfield widget."""
    return Widget(
        text,
        EDITABLE | FOCUSABLE | HAS_BORDER | HAS_BACKGROUND,
        type="textfield",
        **kwargs,
    )


@dataclass
class radio(Widget):  # noqa
    """Create a radio button widget."""

    flags: int = CLICKABLE | FOCUSABLE | HAS_BACKGROUND

    def _draw(
        self,
        surface: pygame.Surface,
        has_focus: bool = False,
        checked: bool = False,
        **style,
    ) -> None:
        if has_focus:
            _draw_widget_focus(surface, self, **style)
        _draw_widget_background(surface, self, **style)
        if self.rect is not None:
            box = self.rect.move_to(
                width=self.rect.height,
                right=self.rect.x - 5,
                y=self.rect.y - self.rect.height * 0.1,
            )
            pygame.draw.rect(surface, style.get("border_color", "white"), box, 1)
            if checked:
                check = pygame.FRect(box).scale_by(0.5)
                pygame.draw.rect(surface, style.get("border_color", "white"), check)
        _draw_widget_text(surface, self, **style)


@dataclass
class IMGUI:
    """Immediate mode GUI state."""

    active: int | None = None
    hot: int | None = None
    focus: int | None = None
    tabindex_prev: int | None = None
    widgets: dict[int, Widget] = field(default_factory=dict)
    events: list[Event] = field(default_factory=list)


@contextlib.contextmanager
def render(ui: IMGUI, surface: pygame.Surface, get_styles=None):
    """Render the GUI."""
    yield get_renderer(ui, surface, get_styles)
    ui.hot = None
    if not pygame.mouse.get_pressed()[0]:
        ui.active = None
    elif ui.active is None:
        ui.active = -1


def get_renderer(ui: IMGUI, surface: pygame.Surface, get_styles=None):
    """Get a rendering function for the GUI."""

    def render_fn(widget: Widget, **style) -> Widget:
        frame_info = inspect.stack()[1]
        widget_id = hash((frame_info.filename, frame_info.lineno, id(widget.value)))
        if callable(get_styles):
            style = get_styles(widget) | style
        with _get_widget(ui, widget_id, widget, **style) as _widget:
            _draw_widget(widget, surface, **style)
        return _widget.triggered

    return render_fn


@contextlib.contextmanager
def _get_widget(ui: IMGUI, widget_id: int, widget: Widget, **kwargs):
    widget.rect = _get_widget_rect(widget, **kwargs)
    widget.triggered = False
    widget.pseudo_classes = set()

    if widget.rect.collidepoint(pygame.mouse.get_pos()):
        ui.hot = widget_id
        widget.pseudo_classes.add("hover")

    if ui.hot == widget_id and pygame.mouse.get_pressed()[0]:
        ui.active = widget_id
        widget.pseudo_classes.add("active")
        if widget.flags & CLICKABLE:
            widget.triggered = True

    if widget.flags & FOCUSABLE and ui.focus is None:
        ui.focus = widget_id
        widget.pseudo_classes.add("focus")

    if ui.focus == widget_id:
        widget.pseudo_classes.add("focus")
        _handle_keys(widget, ui)

    yield widget

    if widget.flags & FOCUSABLE:
        ui.tabindex_prev = widget_id


def _get_widget_rect(widget: Widget, **kwargs) -> pygame.Rect:
    if not widget.rect:
        widget.rect = pygame.Rect(0, 0, 0, 0)
        if isinstance(widget.value, str | list):
            font = kwargs.setdefault(
                "font",
                pygame.font.Font(None, kwargs.get("font_size", 20)),
            )
            widget.rect.size = font.size("".join(widget.value))
        if padding := kwargs.get("padding"):
            widget.rect, _ = add_padding(widget.rect, padding)
    if rect_kwargs := get_rect_attrs(kwargs):
        widget.rect = widget.rect.move_to(**rect_kwargs)
    return widget.rect


def _draw_widget_background(surface: pygame.Surface, widget: Widget, **style) -> None:
    if widget.flags & HAS_BACKGROUND and widget.rect:
        color = style.get(
            "background_color",
            "white" if "active" in widget.pseudo_classes else "black",
        )
        pygame.draw.rect(
            surface,
            color,
            widget.rect,
            width=0,
            border_radius=style.get("border_radius", 0),
        )


def _draw_widget_border(surface: pygame.Surface, widget: Widget, **style) -> None:
    if widget.flags & HAS_BORDER and widget.rect:
        pygame.draw.rect(
            surface,
            style.get("border_color", "white"),
            widget.rect,
            width=style.get("border_width", 1),
            border_radius=style.get("border_radius", 0),
        )


def _draw_widget_focus(surface: pygame.Surface, widget: Widget, **style) -> None:
    if widget.rect:
        pygame.draw.rect(
            surface,
            style.get("focus_border_color", "red"),
            widget.rect.inflate(4, 4),
            width=style.get("focus_border_width", 4),
            border_radius=style.get("focus_border_radius", 0),
        )


def _draw_widget_text(surface: pygame.Surface, widget: Widget, **style) -> None:
    if not widget.rect:
        return
    match widget.value:
        case str() | list():
            text = "".join(widget.value)
        case (str(text), _):
            pass
        case _:
            return
    font = style.setdefault(
        "font",
        pygame.font.Font(None, style.get("font_size", 30)),
    )
    color = style.get("color", "white")
    text_img = font.render(text, True, color)
    rect = text_img.get_rect(center=widget.rect.center)
    match Align[style.get("align", "center")]:
        case Align.left:
            rect.left = widget.rect.left
        case Align.right:
            rect.right = widget.rect.right
        case _:
            rect.centerx = widget.rect.centerx
    match VAlign[style.get("valign", "middle")]:
        case VAlign.top:
            rect.top = widget.rect.top
        case VAlign.bottom:
            rect.bottom = widget.rect.bottom
        case _:
            rect.centery = widget.rect.centery
    _, rect = add_padding(rect, style.get("padding", 0))
    surface.blit(text_img, rect)


def _handle_keys(widget: Widget, ui: IMGUI) -> None:
    editable = widget.flags & EDITABLE
    focus = ui.focus
    focused_widget = ui.widgets.get(focus) if focus is not None else None
    value = widget.value
    value_len = len(widget.value)

    for event in (ev for ev in ui.events if ev.type == KEYDOWN):
        if event.key == K_TAB:
            if focus and focused_widget == widget:
                ui.focus = None
                if event.mod & KMOD_SHIFT:
                    ui.focus = ui.tabindex_prev
            # prevent this event from being processed by other widgets
            event.key = 0
        if editable and event.key == K_BACKSPACE:
            value[:] = value[:-1]
            value_len = len(value)
            widget.triggered = True
        if editable and 32 <= event.key < 127 and value_len < 30:
            value.append(event.unicode)
            widget.triggered = True



"""Entity-Component-System (ECS) implementation."""

from collections.abc import Callable
from collections.abc import Iterable
from functools import wraps
from typing import Any
from typing import Concatenate
from typing import get_type_hints

FilterFn = Callable[[Any], bool]
SystemFn = Callable[Concatenate[Any, ...], None]


def filter_entities(filter_fn: FilterFn) -> Callable[[SystemFn], SystemFn]:
    """Decorator to set filter function for system.

    >>> def has_velocity(entity):
    ...     return hasattr(entity, "velocity")
    >>> @filter_entities(has_velocity)
    ... def apply_velocity(entity):
    ...     entity.position += entity.velocity
    """

    def _filtered_system(system_fn: SystemFn) -> SystemFn:

        @wraps(system_fn)
        def _system_fn(entity: Any, **kwargs) -> None:
            if filter_fn(entity):
                system_fn(entity, **kwargs)

        _system_fn.filtered = True  # type: ignore[attr-defined]
        return _system_fn

    return _filtered_system


def get_ecs_update_fn(systems: list[SystemFn]) -> Callable:
    """Return an update function for ECS."""
    filter_cache: dict[SystemFn, SystemFn] = {}

    def get_filtered_system_fn(system: SystemFn) -> SystemFn:
        if getattr(system, "filtered", False):
            return system

        if callable(filter_ := getattr(system, "filter_entities", None)):
            return filter_entities(filter_)(system)

        entity_type = next(iter(get_type_hints(system).values()))
        return filter_entities(lambda _: isinstance(_, entity_type))(system)

    def ecs_update(entities: Iterable[Any], **kwargs) -> None:
        """Update entities with systems."""
        for system in systems:
            if not (system_fn := filter_cache.get(system)):
                system_fn = filter_cache[system] = get_filtered_system_fn(system)
            for entity in entities:
                system_fn(entity, entities=entities, **kwargs)

    return ecs_update



"""Publish-Subscribe pattern."""

from collections.abc import Callable
from functools import wraps
from typing import Any

CANCEL = object()


def channel(fn: Callable | None = None) -> Callable:
    """
    Returns message channel that can be subscribed to.

    Can be used as an object:
    >>> foo = channel()
    >>> foo.subscribe(lambda x: print(f"subscriber 1: {x}"))
    >>> foo.subscribe(lambda x: print(f"subscriber 2: {x}"))
    >>> foo("bar")
    subscriber 1: bar
    subscriber 2: bar

    Can also be used as a decorator:
    >>> @channel
    ... def foo(x):
    ...     print(f"decorated: {x}")
    >>> foo.subscribe(lambda x: print(f"subscriber: {x}"))
    >>> foo("bar")
    subscriber: bar
    decorated: bar

    And a method decorator:
    >>> class Foo:
    ...     @channel
    ...     def bar(self, x):
    ...         print(f"decorated: {x}")
    >>> foo = Foo()
    >>> foo.bar.subscribe(lambda self, x: print(f"subscriber: {x}"))
    >>> foo.bar("quux")
    subscriber: quux
    decorated: quux

    A message can be cancelled by returning `message.cancel` from a subscriber:
    >>> foo = channel()
    >>> foo.subscribe(lambda x: print(f"subscriber 1: {x}") or foo.cancel)
    >>> foo.subscribe(lambda x: print(f"subscriber 2: {x}"))
    >>> foo("bar")
    subscriber 1: bar
    """

    subscribers: list[Callable] = []

    def subscribe(subscriber: Callable) -> None:
        """Subscribe to the message."""
        subscribers.append(subscriber)

    def wrapper(*args, **kwargs) -> Any:
        """Publish the message."""
        cancel = CANCEL
        for subscriber in subscribers:
            if subscriber(*args, **kwargs) == cancel:
                break
        else:
            if fn:
                fn(*args, **kwargs)

    if fn:
        wrapper = wraps(fn)(wrapper)

    wrapper.subscribe = subscribe  # type: ignore
    wrapper.cancel = CANCEL  # type: ignore

    return wrapper



"""Shake effect for camera or other objects."""

import math
from random import uniform

from pygame import Vector2


def shake(
    dampening: float = 0.3,
    direction: Vector2 | None = None,
    magnitude: float = 20,
    noise: float = 1.0,
):
    """Create a shake effect function."""
    if direction is None:
        direction = Vector2(1, 0)

    _magnitude = magnitude

    def _shake(quotient: float) -> Vector2:
        nonlocal _magnitude

        sine_wave = math.sin(quotient)
        _noise = Vector2(uniform(-1, 1), uniform(-1, 1)) * noise
        offset = (direction * sine_wave + _noise).clamp_magnitude(1) * _magnitude
        _magnitude = max(0, _magnitude - dampening)

        return offset

    return _shake



"""
This module provides a function to create a color gradient surface.
"""

from collections.abc import Callable

import pygame
import pygame.constants
from pygame import Color
from pygame import Surface
from pygame.typing import ColorLike
from pygame.typing import Point


def make_color_gradient(
    size: Point,
    start: ColorLike,
    end: ColorLike,
    vertical: bool = True,
    easing_fn: Callable[[float], float] | None = None,
) -> Surface:
    """
    Create a color gradient surface.
    """

    width, height = int(size[0]), int(size[1])

    if vertical:
        positions = zip([0] * height, range(height), strict=True)
        stripe = Surface((1, height), pygame.constants.SRCALPHA)
        quotient = 1 / height
    else:
        positions = zip(range(width), [0] * width, strict=True)
        stripe = Surface((width, 1), pygame.constants.SRCALPHA)
        quotient = 1 / width

    stripe.fill((0, 0, 0, 0))

    for i, pos in enumerate(positions):
        q = i * quotient
        if easing_fn:
            q = easing_fn(q)
        stripe.set_at(pos, Color(start).lerp(Color(end), q))

    return pygame.transform.scale(stripe, (width, height))



"""
Provides a function to scroll the layers of the parallax effect by a given vector.
"""

from collections.abc import Callable
from collections.abc import Iterable

from pygame.sprite import Sprite
from pygame.typing import Point


def scroll_parallax_layers(
    vector: Point,
    get_layers: Callable[[], Iterable[int]],
    get_sprites_from_layer: Callable[[int], Iterable[Sprite]],
    speeds: dict[int, float],
) -> None:
    """Scroll the layers of the parallax effect by the given vector."""
    vx, vy = vector[0], vector[1]
    layers = get_layers()
    for layer in layers:
        speed = speeds.get(layer, 1)
        lvx, lvy = vx * speed, vy * speed
        for sprite in get_sprites_from_layer(layer):
            if rect := sprite.rect:
                rect.move_ip(lvx, -lvy)
                rect.x %= rect.width
                rect.y %= rect.height



"""Surface utilities for Pygame."""

import pygame
from pygame import Rect
from pygame import Surface
from pygame import Vector2
from pygame.sprite import Sprite

from pygskin import get_rect_attrs


def make_sprite(image: Surface, **kw) -> Sprite:
    """Make a sprite object from a Surface."""
    spr = Sprite()
    if bg_color := kw.get("bg_color"):
        image.fill(bg_color)
    if scale := kw.get("scale"):
        image = pygame.transform.scale_by(image, scale)
    spr.image = image
    spr.rect = image.get_rect(**get_rect_attrs(kw))
    return spr


def rotate_surface(
    surface: Surface,
    angle: float,
    center: Vector2 | None = None,
    offset: Vector2 | None = None,
) -> tuple[Surface, Rect]:
    """Rotate a surface and return the rotated surface and its rect."""
    rotated_surface = pygame.transform.rotate(surface, angle)
    offset = offset.rotate(-angle) if offset else Vector2(0)
    center = center or Vector2()
    rect = rotated_surface.get_rect(center=center + offset)
    return rotated_surface, rect



"""Easing functions."""

import math


def sine_in(x: float) -> float:
    return 1 - math.cos((x * math.pi) / 2)


def sine_out(x: float) -> float:
    return math.sin((x * math.pi) / 2)


def sine_in_out(x: float) -> float:
    return -(math.cos(math.pi * x) - 1) / 2


def cubic_in(x: float) -> float:
    return x**3


def cubic_out(x: float) -> float:
    return 1 - (1 - x) ** 3


def cubic_in_out(x: float) -> float:
    return 4 * x**3 if x < 0.5 else 1 - (-2 * x + 2) ** 3 / 2


def expo_in(x: float) -> float:
    return 0 if x == 0 else 2 ** (10 * x - 10)


def expo_out(x: float) -> float:
    return 1 if x == 1 else 1 - 2 ** (-10 * x)


def expo_in_out(x: float) -> float:
    if x in {0, 1}:
        return x
    return 2 ** (20 * x - 10) / 2 if x < 0.5 else (2 - 2 ** (-20 * x + 10)) / 2


def bounce_in(x: float) -> float:
    return 1 - bounce_out(1 - x)


def bounce_out(x: float) -> float:
    n1 = 7.5625
    d1 = 2.75
    if x < 1 / d1:
        return n1 * x**2
    if x < 2 / d1:
        x -= 1.5 / d1
        return n1 * x**2 + 0.75
    if x < 2.5 / d1:
        x -= 2.25 / d1
        return n1 * x**2 + 0.9375
    x -= 2.625 / d1
    return n1 * x**2 + 0.984375


def bounce_in_out(x: float) -> float:
    if x < 0.5:
        return (1 - bounce_out(1 - 2 * x)) / 2
    return (1 + bounce_out(2 * x - 1)) / 2



"""Vector utilities."""

import math

from pygame import Vector2


def angle_between(v1: Vector2, v2: Vector2) -> float:
    """
    Calculate the angle from v1 to v2
    0 degrees is down and angle increases anti-clockwise, due to pygame's
    inverted y coordinate system.

    >>> points = [(1, 5), (4, 5), (4, 2), (4, -1), (1, -1), (-2, -1), (-2, 2), (-2, 5)]
    >>> [angle_between(Vector2(1, 2), Vector2(b)) for b in points]
    [0.0, 45.0, 90.0, 135.0, 180.0, 225.0, 270.0, 315.0]
    """
    return math.degrees(math.atan2((v2.x - v1.x), (v2.y - v1.y))) % 360



"""Direction enum class for representing directions in a 2D grid."""

from __future__ import annotations

from enum import IntFlag

from pygame import Vector2


class Direction(IntFlag):
    """Direction enum class for representing directions in a 2D grid."""

    UP = 1
    RIGHT = 2
    DOWN = 4
    LEFT = 8

    VERTICAL = 5
    HORIZONTAL = 10

    @property
    def axis(self) -> Direction:
        """Return the axis of the direction."""
        if self in Direction.VERTICAL:
            return Direction.VERTICAL
        if self in Direction.HORIZONTAL:
            return Direction.HORIZONTAL
        return Direction(0)

    @property
    def vector(self) -> Vector2:
        """Return the unit vector of the direction."""
        return {
            self.UP: Vector2(0, -1),
            self.DOWN: Vector2(0, 1),
            self.LEFT: Vector2(-1, 0),
            self.RIGHT: Vector2(1, 0),
        }[self]



"""
A state machine implemented with a coroutine.
"""

from collections.abc import Callable
from collections.abc import Generator
from typing import Any
from typing import TypeVar

RESET = object()
T = TypeVar("T")


def first_key(d: dict[T, Any]) -> T:
    """Return the first key in a dictionary."""
    return next(iter(d))


def statemachine(
    transition_table: dict[Any, list[Callable]],
) -> Generator[Any, Any, None]:
    """
    A state machine implemented with a coroutine.

    >>> s = {"code": [1, 2, 3], "buffer": []}
    >>> def unlock(i):
    ...     buffer = s["buffer"] + [i]
    ...     if buffer == s["code"]:
    ...         return "unlocked"
    >>> def error(i):
    ...     buffer = s["buffer"] + [i]
    ...     if len(buffer) == len(s["code"]) and buffer != s["code"]:
    ...         return lock()
    >>> def digit(i):
    ...     if isinstance(i, int) and 0 <= i <= 9:
    ...         s["buffer"].append(i)
    ...         return "entering_code"
    >>> def lock(*_):
    ...     s["buffer"].clear()
    ...     return "locked"
    >>> safe = statemachine({
    ...     "locked": [digit],
    ...     "entering_code": [unlock, error, digit],
    ...     "unlocked": [lock],
    ... })
    >>> next(safe)
    'locked'
    >>> safe.send('A')
    'locked'
    >>> safe.send(5)
    'entering_code'
    >>> safe.send(5)
    'entering_code'
    >>> safe.send(5)
    'locked'
    >>> safe.send(1)
    'entering_code'
    >>> safe.send(2)
    'entering_code'
    >>> safe.send(3)
    'unlocked'
    """

    reset = RESET
    state = first_key(transition_table)
    while state:
        input_ = yield state
        if input_ is reset:
            state = first_key(transition_table)
            yield None
            continue
        transitions = transition_table[state]
        for transition in transitions:
            if next_state := transition(input_):
                state = next_state
                break


statemachine.RESET = RESET  # type: ignore



"""Module for defining remappable controls."""

from collections.abc import Callable
from functools import partial

from pygame.event import Event


def match_event(event: Event, other: Event) -> bool:
    """Return whether the given events match."""
    return event.type == other.type and event.__dict__.items() >= other.__dict__.items()


def get_action_mapper(mapping: dict[str, Event]) -> Callable[[Event], str | None]:
    """Return a function that maps input events to actions."""

    def get_action(event: Event) -> str | None:
        is_match = partial(match_event, event)
        for action, input_ in mapping.items():
            if input_ and is_match(input_):
                return action
        return None

    return get_action


def map_inputs_to_actions(mapping: dict[str, Event], events: list[Event]) -> list[str]:
    """Return a list of actions from the given events."""
    return list(filter(None, map(get_action_mapper(mapping), events)))



import pygame
import pygame.locals as pg

from pygskin import Assets
from pygskin import imgui
from pygskin import run_game
from pygskin.dialogue import iter_dialogue

assets = Assets()
gui = imgui.IMGUI()


def main():
    text = ["..."]
    choices = []

    def speak(actor, line):
        text[:] = f"{actor}: {line}"

    def stage_direction(params):
        text[:] = f"[{params}]"

    def prompt(options):
        choices[:] = options
        text[:] = "\n".join(
            ". ".join(map(str, pair))
            for pair in enumerate((option["text"] for option in options), 1)
        )

    context = {}
    dialogue = iter_dialogue(
        assets.ray_first_meeting,
        context=context,
        speak=speak,
        stage_direction=stage_direction,
        prompt=prompt,
    )

    def main_loop(surface, events, exit):
        surface.fill("black")

        for event in events:
            if event.type == pg.KEYDOWN:
                is_choice_key = event.key in range(pg.K_1, pg.K_1 + len(choices))

                if choices and is_choice_key:
                    if choice := choices[event.key - pg.K_1]:
                        context["choice"] = choice
                        choices[:] = []

                elif action := next(dialogue, None):
                    action()

        with imgui.render(gui, surface) as render:
            render(
                imgui.label("".join(text)),
                topleft=(20, 20),
                size=(760, 260),
                align="left",
                valign="top",
            )
            render(
                imgui.label("\n".join(str(_) for _ in context.items())),
                topleft=(20, 320),
                size=(760, 260),
                align="left",
                valign="top",
            )

    return main_loop


if __name__ == "__main__":
    run_game(pygame.Window("Dialogue Demo", (800, 600)), main())



"""Rect utility functions."""

from collections.abc import Callable
from collections.abc import Sequence
from typing import overload

from pygame import Rect
from pygame.typing import IntPoint

RECT_ATTRS = """
    x y
    top left bottom right
    topleft bottomleft topright bottomright
    midtop midleft midbottom midright
    center centerx centery
    size width height
    w h
""".strip().split()


def get_rect_attrs(d: dict) -> dict:
    """Return a dictionary with only the keys that are valid Rect attributes."""
    return {k: v for k, v in d.items() if k in RECT_ATTRS}


def add_padding(
    rect: Rect,
    padding: int | Sequence[int] = 0,
) -> tuple[Rect, Rect]:
    """Add padding to a rectangle."""
    match padding:
        case int(px):
            return rect.inflate(px * 2, px * 2), rect.move(px, px)

        case Sequence() as p if 1 <= len(p) <= 4 and all(isinstance(i, int) for i in p):
            try:
                it = iter(p)
                top = right = bottom = left = next(it)
                right = left = next(it)
                bottom = next(it)
                left = next(it)
            except StopIteration:
                pass
            return (
                rect.inflate(left + right, top + bottom).move_to(
                    top=rect.top - top, left=rect.left - left
                ),
                rect.move(left, top),
            )

    raise ValueError("Invalid padding")


def grid(
    rect: Rect,
    rows: int = 1,
    columns: int = 1,
    names: dict[str, IntPoint] | None = None,
) -> Callable[..., Rect]:
    """Return a function to access cells in a grid of rectangles.

    >>> rect = Rect(0, 0, 100, 50)
    >>> get_cell = grid(rect, rows=2, columns=4, names={"foo": (3, 1)})
    >>> get_cell(1, 1)
    Rect(25, 25, 25, 25)
    >>> get_cell("foo")
    Rect(75, 25, 25, 25)
    """

    cellw, cellh = rect.width // columns, rect.height // rows

    @overload
    def get_cell(key: str) -> Rect: ...

    @overload
    def get_cell(column: int, row: int) -> Rect: ...

    def get_cell(*args, **_) -> Rect:
        match args:
            case (str(key),) if names and key in names:
                coord = names[key]
                column, row = coord[0], coord[1]
            case (int(column), int(row)):
                ...
            case _:
                raise KeyError(*args)
        if not (0 <= column < columns and 0 <= row < rows):
            raise KeyError(*args)
        return Rect(column * cellw, row * cellh, cellw, cellh)

    return get_cell



import pygame

from pygskin import Assets
from pygskin import Timer
from pygskin import animate
from pygskin import run_game
from pygskin import spritesheet


def main():
    assets = Assets()
    sheet = spritesheet(pygame.transform.scale_by(assets.cat, 4), rows=3, columns=8)
    walk_left = [sheet(x, 0) for x in range(8)]
    timer = Timer(300, loop=-1)
    anim = animate(walk_left, timer.quotient)

    def main_loop(surface, events, exit):
        timer.tick()
        surface.fill((0, 0, 0))
        surface.blit(next(anim), (160, 160))

    return main_loop


if __name__ == "__main__":
    run_game(pygame.Window("Spritesheet Animation", (400, 400)), main())



"""
Example of customizing keyboard controls.
"""

from functools import partial

from pygame import Event
from pygame.key import name as key_name
from pygame.locals import (
    K_DOWN,
    K_ESCAPE,
    K_LEFT,
    K_RIGHT,
    K_UP,
    KEYDOWN,
)
from pygame.window import Window

from pygskin import imgui
from pygskin import map_inputs_to_actions
from pygskin import run_game
from pygskin.imgui import button, label


DEFAULT_KEY_CONTROLS = {
    "up": Event(KEYDOWN, key=K_UP),
    "down": Event(KEYDOWN, key=K_DOWN),
    "left": Event(KEYDOWN, key=K_LEFT),
    "right": Event(KEYDOWN, key=K_RIGHT),
    "quit": Event(KEYDOWN, key=K_ESCAPE),
}


def main():
    """
    Customizable controls example.
    """
    gui = imgui.IMGUI()
    text = ""
    action_map = DEFAULT_KEY_CONTROLS.copy()
    get_actions = partial(map_inputs_to_actions, action_map)
    waiting_for_input = None

    def main_loop(screen, events, exit_):
        """Test function for the game loop."""
        nonlocal text, waiting_for_input

        screen.fill("black")

        for action in get_actions(events):
            text = action

            if action == "quit":
                exit_()

        if waiting_for_input:
            for event in events:
                if event.type == KEYDOWN:
                    action_map[waiting_for_input] = event
                    waiting_for_input = None
                    break

        with imgui.render(gui, screen) as render:
            render(imgui.label("Set Controls"), font_size=40, center=(400, 100))

            for i, (action, event) in enumerate(action_map.items()):
                if render(button(action), padding=[10], center=(300, 200 + i * 50)):
                    if not waiting_for_input:
                        action_map[action] = None
                        waiting_for_input = action
                if event:
                    render(
                        label(f"{key_name(event.key)}"),
                        center=(500, 200 + i * 50),
                    )

            render(imgui.label(text), font_size=30, center=(400, 500))

    return main_loop


if __name__ == "__main__":
    run_game(Window("Testing", (800, 600)), main())



from pygskin import easing
from pygskin import imgui
from pygskin.animation import EasingFn
from pygskin.animation import LerpFn
from pygskin.animation import animate
from pygskin.assets import Assets
from pygskin.camera import Camera
from pygskin.clock import Clock
from pygskin.clock import Timer
from pygskin.dialogue import iter_dialogue
from pygskin.direction import Direction
from pygskin.ecs import FilterFn
from pygskin.ecs import SystemFn
from pygskin.ecs import filter_entities
from pygskin.ecs import get_ecs_update_fn
from pygskin.func import bind
from pygskin.game import run_game
from pygskin.gradient import make_color_gradient
from pygskin.input import map_inputs_to_actions
from pygskin.lazy import lazy
from pygskin.parallax import scroll_parallax_layers
from pygskin.pubsub import channel
from pygskin.rect import add_padding
from pygskin.rect import get_rect_attrs
from pygskin.rect import grid
from pygskin.screen import ScreenFn
from pygskin.screen import screen_manager
from pygskin.shake import shake
from pygskin.spritesheet import spritesheet
from pygskin.statemachine import statemachine
from pygskin.stylesheet import get_styles
from pygskin.surface import make_sprite
from pygskin.surface import rotate_surface
from pygskin.text import speech_duration
from pygskin.text import to_snakecase
from pygskin.tile import tile
from pygskin.vector import angle_between

__all__ = [
    "Assets",
    "Camera",
    "Clock",
    "Direction",
    "EasingFn",
    "FilterFn",
    "LerpFn",
    "ScreenFn",
    "SystemFn",
    "Timer",
    "add_padding",
    "angle_between",
    "animate",
    "bind",
    "channel",
    "easing",
    "filter_entities",
    "get_ecs_update_fn",
    "get_rect_attrs",
    "get_styles",
    "grid",
    "imgui",
    "iter_dialogue",
    "lazy",
    "make_color_gradient",
    "make_sprite",
    "map_inputs_to_actions",
    "rotate_surface",
    "run_game",
    "screen_manager",
    "scroll_parallax_layers",
    "shake",
    "speech_duration",
    "spritesheet",
    "statemachine",
    "tile",
    "to_snakecase",
]



"""Shake animation example"""

from collections.abc import Callable

import pygame
import pygame.locals as pg
from pygame import Event
from pygame import Rect
from pygame import Surface
from pygame import Window

from pygskin import ScreenFn
from pygskin import Timer
from pygskin import animate
from pygskin import run_game
from pygskin import shake


def shake_example() -> ScreenFn:
    """Shake animation example."""

    rect = Rect(350, 250, 100, 100)
    timer = Timer(3000)
    shake_anim = animate(shake(), timer.quotient)

    def main_loop(screen: Surface, events: list[Event], exit: Callable) -> None:
        nonlocal shake_anim

        screen.fill("black")
        timer.tick()
        pygame.draw.rect(screen, "red", rect.move(next(shake_anim)))

        if any(event.type == pg.KEYDOWN for event in events):
            timer.elapsed = 0
            shake_anim = animate(shake(), timer.quotient)

    return main_loop


if __name__ == "__main__":
    run_game(Window("Camera shake", (800, 600)), shake_example())



"""Access cells in a spritesheet."""

from collections.abc import Callable
from functools import cache

from pygame import Surface

from pygskin.rect import grid


def spritesheet(image: Surface, *args, **kwargs) -> Callable[..., Surface]:
    """Return a function to access cells in a spritesheet.

    >>> image = Surface((20, 10))
    >>> colors = {"r": [255, 0, 0, 255], "g": [0, 255, 0, 255], "b": [0, 0, 255, 255]}
    >>> coords = [(x, 0) for x in range(3)]
    >>> names = dict(zip(colors.keys(), coords))
    >>> for i, (color, (x, y)) in enumerate(names.items()):
    ...     image.fill(colors[color], (x * 5, y * 5, 5, 5))
    Rect(0, 0, 5, 5)
    ...
    >>> sheet = spritesheet(image, rows=1, columns=3, names=names)
    >>> all(list(sheet(name).get_at((0, 0))) == color for name, color in colors.items())
    True
    """

    get_cell = grid(image.get_rect(), *args, **kwargs)

    @cache
    def get_subsurface(*args, **kwargs) -> Surface:
        return image.subsurface(get_cell(*args, **kwargs))

    return get_subsurface



"""A module for managing the game clock and timers."""

from dataclasses import dataclass

import pygame

Clock = pygame.time.Clock()


@dataclass
class Timer:
    """A timer that ticks milliseconds down to zero."""

    duration: int
    elapsed: int = 0
    paused: bool = False
    loop: int | bool = False

    def __hash__(self) -> int:
        return hash(id(self))

    def tick(self, delta_time: int | None = None) -> None:
        """Advance the timer by a given amount of time."""
        if self.paused:
            return

        delta_time = delta_time or Clock.get_time()
        self.elapsed = min(self.elapsed + delta_time, self.duration)

        if self.finished and self.loop:
            self.elapsed = 0
            if self.loop > 0:
                self.loop -= 1

    def quotient(self) -> float:
        """The proportion of time elapsed."""
        return self.elapsed / self.duration

    @property
    def finished(self) -> bool:
        """Whether the timer has finished."""
        return self.elapsed >= self.duration



"""A lazy object."""

from collections.abc import Callable
from functools import cache
from functools import partial
from typing import Any
from typing import Generic
from typing import TypeVar

T = TypeVar("T")


class lazy(Generic[T]):  # noqa: N801
    """A lazy object.

    This class is used to defer loading of an object until it is accessed.

    >>> class Foo:
    ...     def __init__(self):
    ...         print("Foo.__init__ called")
    ...         self.bar = 42
    ...
    >>> lazy_foo = lazy(Foo)
    >>> lazy_foo.bar
    Foo.__init__ called
    42
    >>> lazy_foo.bar
    42
    >>> isinstance(lazy_foo, Foo)
    True
    >>> type(lazy_foo)
    <class 'pygskin.lazy.lazy'>
    """

    def __init__(self, loader: Callable[[], T], *args, **kwargs) -> None:
        self._loader = cache(partial(loader, *args, **kwargs))

    def _load(self) -> T:
        return self._loader()

    def __getattr__(self, name: str) -> Any:
        return getattr(self._load(), name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_load"):
            super().__setattr__(name, value)
        else:
            setattr(self._load(), name, value)

    def __delattr__(self, name: str) -> None:
        if name.startswith("_load"):
            raise AttributeError(f"cannot delete attribute '{name}'")
        delattr(self._load(), name)

    __members__ = property(dir)

    def __dir__(self) -> list[str]:
        return dir(self._load())

    @property
    def __class__(self) -> type:
        return self._load().__class__

    @__class__.setter
    def __class__(self, cls: type) -> None:
        pass



"""Attribute access to static assets in a directory."""

import inspect
import json
from collections import UserDict
from collections.abc import Callable
from collections.abc import Iterator
from functools import cache
from pathlib import Path
from typing import Any

import pygame

type Asset = Any  # type: ignore


def _load_font(path: Path):
    return cache(lambda size=30: pygame.font.Font(path, size))


def _load_json(path: Path):
    return json.loads(path.read_text())


def _load_yaml(path: Path):
    import yaml

    return yaml.safe_load(path.read_text())


LOADERS: dict[str, Callable] = {
    **{_: pygame.image.load for _ in [".png", ".jpg", ".jpeg", ".gif"]},
    **{_: pygame.mixer.Sound for _ in [".wav", ".mp3", ".ogg"]},
    **{_: _load_font for _ in [".ttf"]},
    **{_: _load_yaml for _ in [".yaml", ".yml"]},
    **{_: _load_json for _ in [".json"]},
}


def _get_assets_path_from_caller_frame() -> Path:
    try:
        return Path(inspect.stack()[2].filename).parent / "assets"
    except IndexError:
        return Path("assets")


def _find_assets_by_name(path: Path, name: str) -> Iterator[Path]:
    yield from sorted(
        (m for m in path.glob(f"{name}*.*") if m.suffix in LOADERS),
        # shortest match first
        key=lambda _: len(str(_)),
    )


def _get_asset(path: Path, name: str) -> Asset:
    match next(_find_assets_by_name(path.parent, name), None):
        case Path() as p if p.is_dir():
            asset = Assets(p)
        case Path() as p:
            asset = LOADERS[p.suffix](p)
        case None:
            raise LookupError(f"Asset not found: {path / name}")
    return asset


class Assets(UserDict):
    """
    Provides attribute access to static assets in a directory.

    >>> import tempfile
    >>> tempdir = tempfile.TemporaryDirectory()
    >>> Path(f"{tempdir.name}/foo.gif").write_bytes(
    ...     bytes.fromhex('4749463839610100010000ff002c00000000010001000002003b'))
    26
    >>> _ = pygame.init()
    >>> _ = pygame.display.set_mode((1, 1), pygame.HIDDEN)
    >>> Assets(tempdir.name).foo.get_size()
    (1, 1)
    """

    def __init__(self, path: str | Path | None = None) -> None:
        super().__init__()
        match path:
            case Path():
                ...
            case str(s):
                path = Path(s)
            case None:
                path = _get_assets_path_from_caller_frame()
            case _:
                raise TypeError(f"Expected Path or str, got {type(path)}")
        if not path.is_dir():
            raise ValueError(f"Path does not exist: {path}")
        self.__dict__["path"] = path

    def __getitem__(self, name: str) -> Asset:
        if not (asset := self.data.get(name)):
            try:
                asset = self.data[name] = _get_asset(self.__dict__["path"] / name, name)
            except LookupError as e:
                raise KeyError(name) from e
        return asset

    def __getattr__(self, name: str) -> Asset:
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError from e

    def load_all(self) -> None:
        """Load all assets in the directory into cache."""
        for child in list(self.path.iterdir()):
            _ = self[child.stem]



"""A module for running a game loop with Pygame."""

from asyncio import run
from asyncio import sleep

import pygame
import pygame.constants
from pygame import Window
from pygame.event import get as get_events

from pygskin import Clock


def run_game(window: Window, fn, fps: int = 60):
    """
    Run a game loop with the given function.

    Args:
        window (pygame.Window): The window to run the game in.
        fn (Callable): The function to run each frame.
        fps (int): The frames per second to run the game at.
    """
    pygame.init()
    surface = window.get_surface()

    async def _main_loop():
        running = True
        quit_event_type = pygame.constants.QUIT

        def stop():
            nonlocal running
            running = False

        while running:
            Clock.tick(fps)
            events = get_events()
            for event in events:
                if event.type == quit_event_type:
                    stop()
            fn(surface, events, stop)
            window.flip()
            await sleep(0)

    run(_main_loop())



"""Screen manager for Pygame applications."""

from collections.abc import Callable
from contextlib import suppress
from typing import Any

from pygame import Surface
from pygame.event import Event

from pygskin.statemachine import statemachine

ScreenFn = Callable[[Surface, list[Event], Callable], None]
Transition = Callable[[Any], ScreenFn | None]
TransitionTable = dict[ScreenFn, list[Transition]]
EXIT = object()


def screen_manager(*screens: ScreenFn) -> Callable:
    """Return a screen manager function."""

    def transition_screen(to: ScreenFn) -> ScreenFn | None:
        if to not in screens:
            raise ValueError(f"Transition to unknown screen: {to}")
        return to

    sm = statemachine({screen: [transition_screen] for screen in screens})
    screen = next(sm)

    def exit_manager(*_) -> None:
        raise ValueError("exit_manager not set")

    def exit_screen(to: Transition | None = None) -> None:
        nonlocal screen
        if to is None:
            exit_manager()
            return
        with suppress(StopIteration):
            screen = sm.send(to)

    def _screen_manager(surface: Surface, events: list[Event], exit_: Callable):
        nonlocal exit_manager
        exit_manager = exit_
        screen(surface, events, exit_screen)

    return _screen_manager



from collections.abc import Callable
from contextlib import contextmanager
from functools import partial

import pygame
from pygame.event import Event

from pygskin import imgui
from pygskin.game import run_game
from pygskin.imgui import button
from pygskin.imgui import label
from pygskin.imgui import radio
from pygskin.imgui import textfield
from pygskin.stylesheet import get_styles

gui = imgui.IMGUI()
stylesheet = partial(
    get_styles,
    {
        "label": {
            "color": "green",
            "font_size": 40,
        },
        "textfield": {
            "background_color": "white",
            "color": "black",
            "font_size": 30,
            "align": "left",
        },
        "button:hover": {
            "border_color": "green",
        },
    },
)


@contextmanager
def render_gui(surface: pygame.Surface, events: list[Event]):
    gui.events = events
    with imgui.render(gui, surface, stylesheet) as render:
        yield render


def main() -> Callable[[pygame.Surface, list[Event], Callable], None]:
    foo: list[str] = list("Hello, World!")
    bar: list[str] = list("Hello yourself!")
    choices: dict[str, int] = {
        "One": 1,
        "Two": 2,
        "Three": 3,
    }
    shared: dict = {"choice": 1}

    def _main(surface: pygame.Surface, events: list[Event], exit) -> None:
        surface.fill((0, 0, 0))

        with render_gui(surface, events) as gui:
            gui(label(foo), font_size=40, center=(400, 100))
            gui(textfield(bar), size=(400, 50), center=(400, 200))
            if gui(button("Click me"), size=(200, 50), center=(400, 300)):
                foo[:] = bar[:]

            def option(i: int, text: str, checked: bool):
                return gui(radio(text), checked=checked, x=400, y=400 + 50 * i)

            for i, (text, value) in enumerate(choices.items()):
                if option(i, text, checked=shared["choice"] == value):
                    shared["choice"] = value

    return _main


if __name__ == "__main__":
    run_game(pygame.Window("IMGUI Demo", (800, 600)), main())



import random
from itertools import repeat

import pygame
import pygame.locals as pg
from pygame import Rect
from pygame import Surface
from pygame import Window
from pygame.color import THECOLORS
from pygame.sprite import Group

from pygskin import Assets
from pygskin import Direction
from pygskin import Timer
from pygskin import animate
from pygskin import imgui
from pygskin import make_sprite
from pygskin import run_game
from pygskin import shake
from pygskin import tile
from pygskin.camera import Camera

assets = Assets()
gui = imgui.IMGUI()


def test_camera() -> None:
    """Test the Camera class."""
    camera = Camera()

    world = Surface((1600, 1200), pygame.SRCALPHA)
    for x in range(16):
        for y in range(12):
            pygame.draw.rect(
                world,
                random.choice(list(THECOLORS.values())),
                Rect(x * 100, y * 100, 100, 100),
            )

    direction: Direction | None = None
    zooming: str | None = None
    timer = Timer(3000)
    shake_anim = iter(repeat((0, 0)))
    assets.cat.set_colorkey("black")
    cat = pygame.transform.scale_by(assets.cat.subsurface((0, 0, 16, 16)), 3)
    sprites = Group(
        [
            make_sprite(cat, center=(x * 100, y * 100))
            for x in range(16)
            for y in range(12)
        ]
    )

    def main_loop(screen, events, exit):
        nonlocal direction, shake_anim, zooming

        before_shake = camera.rect.topleft
        camera.view.fill("black")
        camera.rect.move_ip(next(shake_anim))
        camera.blits(tile(camera.rect, world))
        sprites.draw(camera)
        camera.draw(screen)
        camera.rect.topleft = before_shake

        for event in events:
            if event.type == pg.KEYDOWN:
                if event.key == pg.K_UP:
                    direction = Direction.UP
                if event.key == pg.K_DOWN:
                    direction = Direction.DOWN
                if event.key == pg.K_LEFT:
                    direction = Direction.LEFT
                if event.key == pg.K_RIGHT:
                    direction = Direction.RIGHT

                if event.key == pg.K_q:
                    zooming = "in"
                if event.key == pg.K_a:
                    zooming = "out"

                if event.key == pg.K_SPACE:
                    timer.elapsed = 0
                    shake_anim = animate(shake(), timer.quotient)

            if event.type == pg.KEYUP:
                direction = None
                zooming = None

        if zooming == "in":
            camera.zoom += 0.1
        if zooming == "out":
            camera.zoom -= 0.1
        if direction:
            camera.rect.move_ip(direction.vector * (10 / camera.zoom))

        with imgui.render(gui, screen) as render:
            render(
                imgui.label(
                    f"Camera test\n"
                    f"Zoom: {camera.zoom:.2f}\n"
                    f"Position: {camera.rect.topleft}",
                ),
                align="left",
                valign="top",
                topleft=(10, 10),
            )
            render(
                imgui.label(
                    "Arrow keys: move camera, "
                    "Q: zoom in, "
                    "A: zoom out, SPACE: shake",
                ),
                align="left",
                bottomleft=(10, 590),
            )

    return main_loop


if __name__ == "__main__":
    run_game(Window("Camera test", (800, 600)), test_camera())



"""Functional programming utilities."""


def bind(func, *bind_args, **bind_kwargs):
    """Bind arguments to a function."""

    def _bound_fn(*args, **kwargs):
        kwargs = bind_kwargs | kwargs
        iargs = iter(args)
        args = (next(iargs) if arg is ... else arg for arg in bind_args)
        return func(*args, *iargs, **kwargs)

    return _bound_fn



"""
A simple CSS-like stylesheet parser.
"""

import re
from dataclasses import dataclass
from dataclasses import field

SELECTOR = re.compile(r"(\A|[#.[]|::?)(\*|[a-zA-Z0-9-_]+)(?:=([^]]+))?")


@dataclass
class Selector:
    """A stylesheet selector."""

    tag: str | None = None
    id: str | None = None
    classes: list[str] = field(default_factory=list)
    attributes: dict[str, str] = field(default_factory=dict)
    pseudo_classes: list[str] = field(default_factory=list)


def get_styles(stylesheet: dict[str, dict], obj) -> dict:
    """
    Get the styles for an object from a stylesheet.

    >>> from typing import NamedTuple
    >>> Label = NamedTuple("Label", [("id", str), ("classes", list[str])])
    >>> label = Label("foo", ["bar"])
    >>> styles = {
    ...     "Label": {"font_family": "Arial", "font_size": 20, "color": "black"},
    ...     ".bar": {"bold": True, "color": "white"},
    ...     "#foo": {"color": "red"},
    ... }
    >>> get_styles(styles, label)
    {'font_family': 'Arial', 'font_size': 20, 'color': 'red', 'bold': True}
    """
    rulesets = ((parse_selector(k), v) for k, v in stylesheet.items())
    return {
        property: value
        for selector, declarations in sorted(rulesets, key=lambda r: specificity(r[0]))
        for property, value in declarations.items()
        if match(selector, obj)
    }


def parse_selector(s: str) -> Selector:
    """
    Parse a CSS selector.

    >>> from dataclasses import astuple
    >>> astuple(parse_selector("*"))
    ('*', None, [], {}, [])
    >>> astuple(parse_selector("button#submit.btn.btn-primary[type=submit]"))
    ('button', 'submit', ['btn', 'btn-primary'], {'type': 'submit'}, [])
    """
    parts = re.findall(SELECTOR, s)
    selector = Selector()
    for prefix, part, value in parts:
        if prefix == "#":
            selector.id = part
        elif prefix == ".":
            selector.classes.append(part)
        elif prefix == "[":
            key = part
            selector.attributes[key] = value
        elif prefix == ":":
            selector.pseudo_classes.append(part)
        else:
            selector.tag = part
    return selector


def specificity(selector: Selector) -> tuple[int, int, int]:
    """
    Calculate the specificity of a CSS selector.

    >>> specificity((parse_selector("button#submit.btn.btn-primary[type=submit]")))
    (1, 3, 1)
    """
    return (
        1 if selector.id else 0,
        len(selector.classes) + len(selector.attributes) + len(selector.pseudo_classes),
        1 if selector.tag and selector.tag != "*" else 0,
    )


def match(selector: Selector, obj) -> bool:
    """
    Check if the selector matches the given element.

    >>> from typing import NamedTuple
    >>> Image = NamedTuple("Image", [("id", str), ("classes", list[str]), ("src", str)])
    >>> el = Image("foo", ["bar", "red"], "/")
    >>> match(parse_selector("#foo"), el)
    True
    >>> match(parse_selector(".bar"), el)
    True
    >>> match(parse_selector("[src=/]"), el)
    True
    >>> match(parse_selector("Image"), el)
    True
    >>> match(parse_selector("Image#foo.bar.red[src=/]"), el)
    True
    >>> match(parse_selector("Image.bar.blue"), el)
    False
    """
    obj_type = getattr(obj, "type", obj.__class__.__name__)
    return (
        (not selector.tag or selector.tag == "*" or obj_type == selector.tag)
        and (not selector.id or obj.id == selector.id)
        and all(cls in obj.classes for cls in selector.classes)
        and all(
            getattr(obj, key, None) == value
            for key, value in selector.attributes.items()
        )
    )



"""A simple animation system."""

from bisect import bisect
from collections.abc import Callable
from collections.abc import Iterator
from typing import Generic
from typing import Protocol
from typing import TypeVar
from typing import cast
from typing import runtime_checkable

from pygame.math import clamp

T = TypeVar("T")


EasingFn = Callable[[float], float]
LerpFn = Callable[[T, T, float], T]


@runtime_checkable
class Lerpable(Generic[T], Protocol):
    """Can be plugged into standard linear interpolation algorithm."""

    def __add__(self: T, other: T) -> T: ...

    def __mul__(self: T, other: float) -> T: ...

    def __sub__(self: T, other: T) -> T: ...


def lerp(start, end, quotient):
    """Linear interpolation between two values."""
    if isinstance(start, dict) and isinstance(end, dict):
        return {key: lerp(value, end[key], quotient) for key, value in start.items()}

    if hasattr(start, "lerp") and callable(start.lerp):
        return start.lerp(end, quotient)

    if isinstance(start, Lerpable):
        return start + ((end - start) * quotient)

    # can't interpolate
    return start


def get_keyframe_interpolator(
    keyframes: dict[float, T] | list[T],
    lerp_fn: LerpFn = lerp,
) -> Callable[[float], T]:
    """Get a keyframe interpolator function.

    >>> frames = [2.0, 8.0, 3.0]
    >>> from pygskin import Timer
    >>> timer = Timer(3000)
    >>> anim = animate(frames, timer.quotient)
    >>> next(anim)
    2.0
    >>> timer.tick(1500)
    >>> next(anim)
    8.0
    >>> timer.tick(750)
    >>> next(anim)
    5.5
    >>> timer.tick(750)
    >>> next(anim)
    3.0
    """

    match keyframes:

        case list():
            num = len(keyframes)
            if num < 2:
                raise ValueError("At least two frames are required")
            keyframes = {i / (num - 1): keyframes[i] for i in range(num)}

        case dict():
            pass

        case _:
            raise TypeError("Invalid keyframes")

    keys = sorted(keyframes.keys())
    if min(keys) < 0.0 or max(keys) > 1.0:
        raise ValueError("Keyframes must span [0, 1]")
    first_key = keys[0]
    last_key = keys[-1]
    last_key_pos = len(keys) - 1

    def get_frame_at(key):
        easing = None
        key_pos = bisect(keys, key)
        start_key = keys[max(0, key_pos - 1)]
        if isinstance(start_frame := keyframes[start_key], tuple):
            start_frame, easing = start_frame

        if start_key == last_key:
            return cast(T, start_frame)

        end_key = keys[min(last_key_pos, key_pos)]
        if isinstance(end_frame := keyframes[end_key], tuple):
            end_frame, _ = end_frame

        if end_key == first_key:
            return cast(T, end_frame)

        duration = end_key - start_key
        quotient = (key - start_key) / duration

        if easing:
            quotient = easing(quotient)

        return lerp_fn(start_frame, end_frame, quotient)

    return get_frame_at


def animate(
    frames: Callable[[float], T] | dict[float, T] | list[T],
    get_quotient: Callable[[], float],
) -> Iterator[T]:
    """
    Animate a sequence of frames.

    >>> import math
    >>> from pygskin import Timer
    >>> timer = Timer(3000)
    >>> anim = animate(lambda q: math.sin(q * math.pi), timer.quotient)
    >>> next(anim)
    0.0
    >>> timer.tick(1500)
    >>> next(anim)
    1.0
    >>> timer.tick(750)
    >>> next(anim)
    0.7071067811865476
    >>> timer.tick(750)
    >>> next(anim)
    1.2246467991473532e-16
    """
    match frames:
        case dict() | list():
            get_frame_at = get_keyframe_interpolator(frames)
        case Callable():
            get_frame_at = frames
        case _:
            raise TypeError("Invalid frames")

    while True:
        yield get_frame_at(clamp(get_quotient(), 0.0, 1.0))



"""A 2D camera class."""

from collections.abc import Iterable
from functools import cached_property

import pygame
from pygame import Rect
from pygame import Surface
from pygame import Vector2
from pygame.math import clamp
from pygame.typing import Point
from pygame.typing import RectLike


class Camera:
    """A 2D camera."""

    def __init__(self, rect: Rect | None = None) -> None:
        self.rect = rect or Rect(0, 0, 800, 600)
        self._view: Surface | None = None
        self._zoom = 1.0
        self._original_size = Vector2(self.rect.size)

    @cached_property
    def view(self) -> Surface:
        surface = Surface(self._original_size * (1 / self.zoom))
        self.rect = surface.get_rect(center=self.rect.center)
        return surface

    @property
    def zoom(self) -> float:
        return self._zoom

    @zoom.setter
    def zoom(self, value: float) -> None:
        self._zoom = clamp(value, 0.1, 10)
        del self.view
        _ = self.view

    def blit(self, source: Surface, dest: Point | None = None) -> None:
        if dest is None:
            dest = Vector2()
        self.view.blit(source, Vector2(dest[:2]) - self.rect.topleft)

    def blits(self, blit_sequence: Iterable[tuple[Surface, Point | RectLike]]) -> None:
        blit_seq = [
            (source, Vector2(dest[:2]) - self.rect.topleft)
            for source, dest in blit_sequence
        ]
        return self.view.blits(blit_seq)

    def draw(self, surface: Surface) -> None:
        surface.blit(pygame.transform.scale_by(self.view, self.zoom))



"""Example of a simple game with a main menu and a play screen."""

import random
from collections.abc import Callable
from functools import cache

import pygame
import pygame.locals as pg
from pygame import Color
from pygame import FRect
from pygame import Surface
from pygame import Vector2
from pygame.color import THECOLORS
from pygame.event import Event
from pygame.sprite import Group
from pygame.sprite import Sprite

from pygskin import get_ecs_update_fn
from pygskin import imgui
from pygskin import run_game
from pygskin import screen_manager
from pygskin.imgui import button
from pygskin.imgui import label
from pygskin.imgui import textfield

gui = imgui.IMGUI()
shared: dict = {"buffer": list("1000"), "num_balls": "1000"}


def main():
    """
    Return a screen manager for the main menu and game screens.
    """
    return screen_manager(
        main_menu,
        play_game(),
    )


def main_menu(surface: Surface, events: list[Event], exit_screen: Callable) -> None:
    """
    Display the main menu screen.
    """
    surface.fill((0, 0, 0))
    gui.events = events
    with imgui.render(gui, surface) as render:
        render(label("Main Menu"), font_size=40, center=(400, 100))

        render(textfield(shared["buffer"]), size=(200, 50), center=(400, 200))

        if render(button("Start Game"), size=(200, 50), center=(400, 300)):
            shared["num_balls"] = int("".join(shared["buffer"]))
            exit_screen(to=play_game())


class Ball(Sprite):
    """
    A simple ball sprite.
    """

    def __init__(self) -> None:
        super().__init__()
        self.vel = Vector2(random.uniform(-6, 6), random.uniform(-6, 6))
        self.image = Surface((10, 10)).convert_alpha()
        self.image.fill((0, 0, 0, 0))
        pygame.draw.circle(
            self.image,
            Color(random.choice(list(THECOLORS.keys()))),
            (5, 5),
            5,
        )
        self.rect = FRect(
            self.image.get_rect(
                centerx=random.randint(0, 800),
                centery=random.randint(0, 600),
            )
        )


def move(ball: Ball, *_, **__) -> None:
    """
    Move the ball sprite.
    """
    if ball.rect is not None:
        new_pos = ball.rect.topleft + ball.vel
        ball.rect.x = new_pos.x % 800.0
        ball.rect.y = new_pos.y % 600.0


@cache
def play_game():
    """
    Return the play game screen function.
    """
    ecs_update = get_ecs_update_fn([move])
    entities = Group()

    def _play(surface: Surface, events: list[Event], exit_screen: Callable) -> None:
        if not entities:
            entities.add(*(Ball() for _ in range(shared["num_balls"])))

        ecs_update(entities)

        surface.fill((0, 0, 0))
        entities.draw(surface)

        for event in events:
            if event.type == pg.KEYDOWN:
                if event.key == pg.K_ESCAPE:
                    exit_screen()

                else:
                    exit_screen(to=main_menu)

    return _play


if __name__ == "__main__":
    run_game(pygame.Window("Testing", (800, 600)), main())



"""Tile an image across a rect."""

from itertools import product

from pygame import Rect
from pygame import Surface
from pygame.sprite import Sprite
from pygame.typing import Point


def tile(
    rect: Rect,
    tile_image: Surface | Sprite,
) -> list[tuple[Surface, Point]]:
    """Get a Surface.blits blit_sequence for tiling an image across the given rect.

    >>> [offset for image, offset in tile(Rect(0, 0, 64, 64), Surface((32, 32)))]
    [(0, 0), (0, 32), (32, 0), (32, 32)]
    """
    match tile_image:
        case Sprite(image=image, rect=tile_rect):
            if not image:
                raise ValueError("tile_image Sprite must have an image attribute")
            if not tile_rect:
                raise ValueError("tile_image Sprite must have a rect attribute")
        case Surface() as image:
            tile_rect = image.get_rect()
        case _:
            raise TypeError("tile_image must be a Surface or Sprite")

    assert tile_rect is not None
    w, h = tile_rect.size
    offsets = (tile_rect.move(x * w, y * h) for x, y in product((-1, 0, 1), repeat=2))
    return [(image, offset.topleft) for offset in offsets if rect.colliderect(offset)]



"""A module for parsing and running dialogue scripts."""

from collections import Counter
from collections import defaultdict
from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Iterator
from functools import partial
from itertools import pairwise

from pygskin.statemachine import statemachine

Action = Callable[[], None]
Dialogue = dict[str, list[dict]]


def iter_dialogue(dialogue: Dialogue, context: dict, **callbacks) -> Iterator[Action]:
    """
    Yields callable actions representing steps in a dialogue script.

    This generator function processes a dialogue graph represented as a parsed
    JSON object.  Each yielded callable performs an action, such as delivering a
    line of dialogue, triggering an event, or prompting user input to determine
    the next branch of the dialogue.

    Args:
        dialogue: A dict representing the parsed dialogue script.
        context: A dict of shared state that persists across the dialogue run.

    Yields:
        A function that performs the next step in the dialogue sequence.
    """
    nodes: dict[str, Generator] = {}

    def callback(name):
        return callbacks.get(name, lambda *_: None)

    def change_node(_):
        return nodes.get(context.pop("next_node", None), None)

    def eval_expr(expr, extra=None):
        return eval(str(expr), context.copy(), extra or {})

    def jump(target=None, condition=None):

        def transition(_, target=target):
            return target if condition is None or eval_expr(condition) else None

        return transition

    def update_context(assignments: dict):
        context.update({k: eval_expr(v) for k, v in assignments.items()})

    def prompt(all_options: list[dict]):
        times_seen: Counter[tuple] = Counter()

        def is_shown(option):
            seen = times_seen[tuple(option.items())]
            return eval_expr(option.get("if", "True"), extra={"_seen": seen})

        def get_options():
            options = [option for option in all_options if is_shown(option)]
            choice = context.pop("choice", None)

            if choice in options:
                context["next_node"] = choice["value"]
                return

            times_seen.update([tuple(option.items()) for option in options])
            callback("prompt")(options)

        return get_options

    def make_action(action_data):
        match action_data:
            case "update_context", dict(assignments):
                return partial(update_context, assignments)
            case "next_node", str(node_name):
                return partial(context.update, next_node=node_name)
            case "pause", float(duration):
                return partial(callback("pause"), duration)
            case "options", list(options):
                return prompt(options)
            case "stage_direction", dict(params):
                return partial(callback("stage_direction"), params)
            case str(actor), str(line):
                return partial(callback("speak"), actor, line)
            case _:
                raise ValueError(f"Invalid action: {action_data}")

    def make_node(name: str, node_data: list[dict]):
        transitions: dict[Action, list[Callable]] = defaultdict(list)
        fork: Action | None = None
        jump_else = jump(None)

        actions = [(_.pop("if", None), make_action(_.popitem())) for _ in node_data]
        actions += [(None, None)]

        # XXX what if the first action has a condition?
        for (_, action), (condition, next_) in pairwise(actions):
            if condition:
                fork = fork or action
                transitions[fork].append(jump(next_, condition))
                if action != fork:
                    transitions[action].append(jump_else)
            else:
                if fork:
                    jump_else.__defaults__ = (next_,)
                    jump_else = jump(None)
                    fork = None
                transitions[action].append(jump(next_, condition))

        nodes[name] = statemachine(transitions)
        return nodes[name], [change_node]

    sm = statemachine(dict(make_node(name, node) for name, node in dialogue.items()))
    node = next(sm)
    input_ = None
    while node:
        action = node.send(input_)
        yield action
        input_ = context
        if "next_node" in context:
            if context["next_node"] == "end":
                break
            node.send(statemachine.RESET)  # type: ignore
            node = sm.send(context)
            input_ = None
